﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;
using System.ServiceModel;
using System.Text;
using System.Data;

namespace NorthwindService
{
    public class NorthwindService : INorthwindService
    {
        public IEnumerable<Product> GetProducts()
        {
            using (var ctx = new NorthwindEntities())
            {
                return ctx.Products.ToList();
            }
        }

        public Customer GetCustomer(string id)
        {
            using (var ctx = new NorthwindEntities())
            {
                return ctx.Customers.Include("Orders")
                    .Where(c => c.CustomerID == id)
                    .SingleOrDefault();
            }
        }

        public bool SubmitOrder(Order newOrder)
        {
            using (var ctx = new NorthwindEntities())
            {
                ctx.Orders.ApplyChanges(newOrder);
                ValidateOrderGraph(ctx, newOrder);
                return ctx.SaveChanges() > 0;
            }
        }

        private void ValidateOrderGraph(NorthwindEntities ctx, Order newOrder)
        {
            ValidationRule(ctx.ObjectStateManager.GetObjectStateEntry(newOrder).State == EntityState.Added, "The submitted order must be in added state.");

            foreach (var modifiedEntry in ctx.ObjectStateManager.GetObjectStateEntries(EntityState.Modified))
            {
                ValidationRule(modifiedEntry.Entity == newOrder.Customer || 
                    (modifiedEntry.Entity is Order  && ((Order) modifiedEntry.Entity).Customer == null), 
                    "Only modified entities allowed are the customer for the submitted order or an order with customer set to null.");
            }

            foreach (var entry in from e in ctx.ObjectStateManager.GetObjectStateEntries(EntityState.Added | EntityState.Deleted)
                                  where e.IsRelationship == false
                                  select e)
            {
                ValidationRule(entry.Entity is Order || entry.Entity is Order_Detail, "Only Orders and Order_Details may be added or deleted.");
                var order = entry.Entity as Order;
                if (order != null && entry.State == EntityState.Deleted)
                {
                    ValidationRule(order.OrderDate == null, "Only Orders with null OrderDates can be deleted.");
                }

                var order_detail = entry.Entity as Order_Detail;
                if (order_detail != null)
                {
                    ValidationRule(order_detail.Order != null, "Order_Detail must have an order.");
                    ValidationRule(entry.State == ctx.ObjectStateManager.GetObjectStateEntry(order_detail.Order).State, "Order_Detail must be in same state as its Order.");
                }
            }
        }

        private void ValidationRule(bool condition, string message)
        {
            if (!condition)
            {
                throw new FaultException<InvalidOperationException>(new InvalidOperationException(message), message);
            }
        }

        public bool SubmitOrderDTO(NewOrderDTO newOrderDTO)
        {
            using (var ctx = new NorthwindEntities())
            {
                var customer = new Customer
                {
                    CustomerID = newOrderDTO.CustomerID,
                    ContactName = newOrderDTO.ContactName,
                    Version = newOrderDTO.CustomerVersion,
                };

                var order = new Order
                {
                    Customer = customer,
                };

                foreach (var line in newOrderDTO.Lines)
                {
                    order.Order_Details.Add(new Order_Detail
                    {
                        ProductID = line.ProductID,
                        Quantity = line.Quantity,
                    });
                }

                ctx.Customers.Attach(customer); // attaches the whole graph
                var customerEntry = ctx.ObjectStateManager.GetObjectStateEntry(customer);
                customerEntry.SetModified(); // only want the one property to be modified
                customerEntry.SetModifiedProperty("ContactName");

                ctx.ObjectStateManager.ChangeObjectState(order, EntityState.Added);
                foreach (var order_detail in order.Order_Details)
                {
                    ctx.ObjectStateManager.ChangeObjectState(order_detail, EntityState.Added);
                }

                return ctx.SaveChanges() > 0;
            }
        }

        public bool UpdateOrderDTO(UpdateOrderDTO updateOrderDTO)
        {
            using (var ctx = new NorthwindEntities())
            {
                var order = new Order
                {
                    OrderDate = updateOrderDTO.OrderDate,
                    Version = updateOrderDTO.Version,
                };
                ctx.Orders.Attach(order);
                var orderEntry = ctx.ObjectStateManager.GetObjectStateEntry(order);
                orderEntry.SetModified();
                orderEntry.SetModifiedProperty("OrderDate");

                foreach (var line in updateOrderDTO.Lines)
                {
                    var detail = new Order_Detail
                    {
                        ProductID = line.ProductID,
                        Quantity = line.Quantity,
                        Version = line.Version,
                    };
                    ctx.Order_Details.Attach(detail);
                    var detailEntry = ctx.ObjectStateManager.GetObjectStateEntry(detail);
                    detailEntry.SetModified();
                    detailEntry.SetModifiedProperty("ProductID");
                    detailEntry.SetModifiedProperty("Quantity");
                    order.Order_Details.Add(detail);
                }

                return ctx.SaveChanges() > 0;
            }
        }
    }
}
